#include "Assembler.hpp"

#include "Token.hpp"
#include "Scanner.hpp"
#include "Parser.hpp"
#include "Linker.hpp"
#include "Loader.hpp"
#include "Error.hpp"
#include "String.hpp"
#include "Operand.hpp"
#include "Synthesizer.hpp"
#include "instructionSet.hpp"

namespace SoftWire
{
	Assembler::Assembler()
	{
		try
		{
			errors = new char[1];
			errors[0] = '\0';

			entryPoint = 0;

			linker = new Linker();
			loader = new Loader(*linker);
			instructionSet = new InstructionSet();
			synthesizer = new Synthesizer();
			scanner = 0;
			parser = 0;
		}
		catch(const Error &error)
		{
			handleError(error.getString());
		}
	}

	Assembler::Assembler(const char *sourceFile)
	{
		try
		{
			errors = new char[1];
			errors[0] = '\0';

			if(!sourceFile)
			{
				throw Error("Input file not specified");
			}

			entryPoint = strdup(sourceFile);
			char *dot = strrchr(entryPoint, '.');

			if(dot)
			{
				*dot = '\0';
			}
			else
			{
				throw Error("Input file must have extension");
			}

			linker = new Linker();
			loader = new Loader(*linker);
			scanner = new Scanner();
			synthesizer = new Synthesizer();
			instructionSet = new InstructionSet();
			parser = new Parser(*scanner, *synthesizer, *instructionSet);

			scanner->scanFile(sourceFile);
			assembleFile();
		}
		catch(const Error &error)
		{
			handleError(error.getString());
		}
	}

	Assembler::~Assembler()
	{
		delete entryPoint;
		entryPoint = 0;

		delete linker;
		linker = 0;

		delete loader;
		loader = 0;

		delete scanner;
		scanner = 0;

		delete synthesizer;
		synthesizer = 0;

		delete instructionSet;
		instructionSet = 0;

		delete parser;
		parser = 0;

		delete[] errors;
		errors = 0;
	}

	void Assembler::assembleFile()
	{
		if(!scanner)
		{
			throw INTERNAL_ERROR;
		}

		scanner->rewind();

		while(!scanner->isEndOfFile())
		{
			assembleLine();

			if(scanner->isEndOfLine() && !scanner->isEndOfFile())
			{
				scanner->advance();
			}
		}

		delete scanner;
		scanner = 0;

		delete parser;
		parser = 0;
	}

	void Assembler::defineExternal(void *pointer, const char *name)
	{
		Linker::defineExternal(pointer, name);
	}

	void Assembler::defineSymbol(int value, const char *name)
	{
		Scanner::defineSymbol(value, name);
	}

	const void (*Assembler::callable(const char *entryPoint))()
	{
		if(!loader)
		{
			throw INTERNAL_ERROR;
		}

		if(errors[0] != '\0')
		{
			return 0;
		}
		else
		{
			if(entryPoint)
			{
				return loader->callable(entryPoint);
			}
			else
			{
				return loader->callable(this->entryPoint);
			}
		}
	}

	void Assembler::assembleLine()
	{
		if(!parser || !loader)
		{
			throw INTERNAL_ERROR;
		}

		const Encoding &encoding = parser->parseLine();
		loader->appendEncoding(encoding);
	}

	const char *Assembler::getErrors() const
	{
		return errors;
	}

	void Assembler::handleError(const char *error)
	{
		const char *sourceLine = parser->skipLine();

		int previousLength = 0;

		if(errors)
		{
			previousLength = strlen(errors);
		}

		int newLength = previousLength + strlen(error) + strlen(sourceLine) + 16;

		delete[] errors;
		errors = new char[newLength];

		strcpy(errors, "error: ");
		strcat(errors, error);
		strcat(errors, "\n\t");
		strcat(errors, sourceLine);
		strcat(errors, "\n");
	}

	const char *Assembler::getListing() const
	{
		return loader->getListing();
	}

	void Assembler::clearListing() const
	{
		loader->clearListing();
	}

	void Assembler::x86(int instructionID, const Operand &firstOperand, const Operand &secondOperand, const Operand &thirdOperand)
	{
		try
		{
			if(!loader || !synthesizer || !instructionSet)
			{
				throw INTERNAL_ERROR;
			}

			const Instruction *instruction = instructionSet->instruction(instructionID);

			synthesizer->reset();

			synthesizer->encodeFirstOperand(firstOperand);
			synthesizer->encodeSecondOperand(secondOperand);
			synthesizer->encodeThirdOperand(thirdOperand);
			const Encoding &encoding = synthesizer->encodeInstruction(instruction);

			loader->appendEncoding(encoding);
		}
		catch(Error &error)
		{
			handleError(error.getString());
		}
	}

	void Assembler::label(const char *label)
	{
		try
		{
			if(!loader || !synthesizer)
			{
				throw INTERNAL_ERROR;
			}

			synthesizer->reset();

			synthesizer->defineLabel(label);
			const Encoding &encoding = synthesizer->encodeInstruction(0);

			loader->appendEncoding(encoding);
		}
		catch(Error &error)
		{
			handleError(error.getString());
		}
	}
};
